@@ -3,15 +3,28 @@
 # Distributed under the terms of the Modified BSD License.
 
 import copy
+import json
+import os
+import platform
 import time
 import warnings
+import socket
+import sys
+
+from pytest import mark
 
 import zmq
 from zmq.tests import (
-    BaseZMQTestCase, SkipTest, have_gevent, GreenTest, skip_pypy, skip_if
+    BaseZMQTestCase, SkipTest, have_gevent, GreenTest, skip_pypy
 )
-from zmq.utils.strtypes import bytes, unicode
+from zmq.utils.strtypes import unicode
 
+pypy = platform.python_implementation().lower() == 'pypy'
+windows = platform.platform().lower().startswith('windows')
+on_travis = bool(os.environ.get('TRAVIS_PYTHON_VERSION'))
+
+# polling on windows is slow
+POLL_TIMEOUT = 1000 if windows else 100
 
 class TestSocket(BaseZMQTestCase):
 
@@ -143,6 +156,8 @@ class TestSocket(BaseZMQTestCase):
                 'ROUTER', 'XPUB', 'TCP', 'FAIL',
                 'REQ_', 'CURVE_', 'PROBE_ROUTER',
                 'IPC_FILTER', 'GSSAPI', 'STREAM_',
+                'VMCI_BUFFER_SIZE', 'VMCI_BUFFER_MIN_SIZE',
+                'VMCI_BUFFER_MAX_SIZE', 'VMCI_CONNECT_TIMEOUT',
                 )):
                 # some sockopts are write-only
                 continue
@@ -200,57 +215,65 @@ class TestSocket(BaseZMQTestCase):
         self.sockets.extend([a,b])
         self.assertRaises(TypeError, a.send_multipart, [b'a', 5])
         a.send_multipart([b'b'])
-        rcvd = b.recv_multipart()
+        rcvd = self.recv_multipart(b)
         self.assertEqual(rcvd, [b'b'])
     
     @skip_pypy
     def test_tracker(self):
         "test the MessageTracker object for tracking when zmq is done with a buffer"
         addr = 'tcp://127.0.0.1'
-        a = self.context.socket(zmq.PUB)
-        port = a.bind_to_random_port(addr)
-        a.close()
-        iface = "%s:%i"%(addr,port)
-        a = self.context.socket(zmq.PAIR)
-        # a.setsockopt(zmq.IDENTITY, b"a")
-        b = self.context.socket(zmq.PAIR)
+        # get a port:
+        sock = socket.socket()
+        sock.bind(('127.0.0.1', 0))
+        port = sock.getsockname()[1]
+        iface = "%s:%i" % (addr, port)
+        sock.close()
+        time.sleep(0.1)
+
+        a = self.context.socket(zmq.PUSH)
+        b = self.context.socket(zmq.PULL)
         self.sockets.extend([a,b])
         a.connect(iface)
         time.sleep(0.1)
         p1 = a.send(b'something', copy=False, track=True)
-        self.assertTrue(isinstance(p1, zmq.MessageTracker))
-        self.assertFalse(p1.done)
+        assert isinstance(p1, zmq.MessageTracker)
+        assert p1 is zmq._FINISHED_TRACKER
+        # small message, should start done
+        assert p1.done
+
+        # disable zero-copy threshold
+        a.copy_threshold = 0
+
         p2 = a.send_multipart([b'something', b'else'], copy=False, track=True)
-        self.assert_(isinstance(p2, zmq.MessageTracker))
-        self.assertEqual(p2.done, False)
-        self.assertEqual(p1.done, False)
+        assert isinstance(p2, zmq.MessageTracker)
+        assert not p2.done
 
         b.bind(iface)
-        msg = b.recv_multipart()
+        msg = self.recv_multipart(b)
         for i in range(10):
             if p1.done:
                 break
             time.sleep(0.1)
         self.assertEqual(p1.done, True)
         self.assertEqual(msg, [b'something'])
-        msg = b.recv_multipart()
+        msg = self.recv_multipart(b)
         for i in range(10):
             if p2.done:
                 break
             time.sleep(0.1)
         self.assertEqual(p2.done, True)
         self.assertEqual(msg, [b'something', b'else'])
-        m = zmq.Frame(b"again", track=True)
+        m = zmq.Frame(b"again", copy=False, track=True)
         self.assertEqual(m.tracker.done, False)
         p1 = a.send(m, copy=False)
         p2 = a.send(m, copy=False)
         self.assertEqual(m.tracker.done, False)
         self.assertEqual(p1.done, False)
         self.assertEqual(p2.done, False)
-        msg = b.recv_multipart()
+        msg = self.recv_multipart(b)
         self.assertEqual(m.tracker.done, False)
         self.assertEqual(msg, [b'again'])
-        msg = b.recv_multipart()
+        msg = self.recv_multipart(b)
         self.assertEqual(m.tracker.done, False)
         self.assertEqual(msg, [b'again'])
         self.assertEqual(p1.done, False)
@@ -265,7 +288,6 @@ class TestSocket(BaseZMQTestCase):
         self.assertEqual(p2.done, True)
         m = zmq.Frame(b'something', track=False)
         self.assertRaises(ValueError, a.send, m, copy=False, track=True)
-        
 
     def test_close(self):
         ctx = self.Context()
@@ -327,7 +349,7 @@ class TestSocket(BaseZMQTestCase):
             a.send(msg)
         time.sleep(0.1)
         for i in range(3):
-            self.assertEqual(b.recv_multipart(), [msg])
+            self.assertEqual(self.recv_multipart(b), [msg])
     
     def test_close_after_destroy(self):
         """s.close() after ctx.destroy() should be fine"""
@@ -342,16 +364,16 @@ class TestSocket(BaseZMQTestCase):
     def test_poll(self):
         a,b = self.create_bound_pair()
         tic = time.time()
-        evt = a.poll(50)
+        evt = a.poll(POLL_TIMEOUT)
         self.assertEqual(evt, 0)
-        evt = a.poll(50, zmq.POLLOUT)
+        evt = a.poll(POLL_TIMEOUT, zmq.POLLOUT)
         self.assertEqual(evt, zmq.POLLOUT)
         msg = b'hi'
         a.send(msg)
-        evt = b.poll(50)
+        evt = b.poll(POLL_TIMEOUT)
         self.assertEqual(evt, zmq.POLLIN)
         msg2 = self.recv(b)
-        evt = b.poll(50)
+        evt = b.poll(POLL_TIMEOUT)
         self.assertEqual(evt, 0)
         self.assertEqual(msg2, msg)
     
@@ -433,9 +455,84 @@ class TestSocket(BaseZMQTestCase):
         a.send(b'hi')
         rcvd = self.recv(b)
         self.assertEqual(rcvd, b'hi')
+    
+    def test_subscribe_method(self):
+        pub, sub = self.create_bound_pair(zmq.PUB, zmq.SUB)
+        sub.subscribe('prefix')
+        sub.subscribe = 'c'
+        p = zmq.Poller()
+        p.register(sub, zmq.POLLIN)
+        # wait for subscription handshake
+        for i in range(100):
+            pub.send(b'canary')
+            events = p.poll(250)
+            if events:
+                break
+        self.recv(sub)
+        pub.send(b'prefixmessage')
+        msg = self.recv(sub)
+        self.assertEqual(msg, b'prefixmessage')
+        sub.unsubscribe('prefix')
+        pub.send(b'prefixmessage')
+        events = p.poll(1000)
+        self.assertEqual(events, [])
+    
+    # Travis can't handle how much memory PyPy uses on this test
+    @mark.skipif(
+        (
+            pypy and on_travis
+        ) or (
+            sys.maxsize < 2**32
+        ) or (
+            windows
+        ),
+        reason="only run on 64b and not on Travis."
+    )
+    def test_large_send(self):
+        try:
+            buf = os.urandom(1) * (2**31 + 1)
+        except MemoryError:
+            raise SkipTest()
+        a, b = self.create_bound_pair()
+        a.send(buf, copy=False)
+        rcvd = b.recv()
+        assert rcvd == buf
+    
+    def test_custom_serialize(self):
+        a, b = self.create_bound_pair(zmq.DEALER, zmq.ROUTER)
+        def serialize(msg):
+            frames = []
+            frames.extend(msg.get('identities', []))
+            content = json.dumps(msg['content']).encode('utf8')
+            frames.append(content)
+            return frames
+        
+        def deserialize(frames):
+            identities = frames[:-1]
+            content = json.loads(frames[-1].decode('utf8'))
+            return {
+                'identities': identities,
+                'content': content,
+            }
+        
+        msg = {
+            'content': {
+                'a': 5,
+                'b': 'bee',
+            }
+        }
+        a.send_serialized(msg, serialize)
+        recvd = b.recv_serialized(deserialize)
+        assert recvd['content'] == msg['content']
+        assert recvd['identities']
+        # bounce back, tests identities
+        b.send_serialized(recvd, serialize)
+        r2 = a.recv_serialized(deserialize)
+        assert r2['content'] == msg['content']
+        assert not r2['identities']
 
 
-if have_gevent:
+if have_gevent and not windows:
     import gevent
     
     class TestSocketGreen(GreenTest, TestSocket):
@@ -450,7 +547,7 @@ if have_gevent:
             self.assertRaises(gevent.Timeout, b.recv)
             g.kill()
         
-        @skip_if(not hasattr(zmq, 'RCVTIMEO'))
+        @mark.skipif(not hasattr(zmq, 'RCVTIMEO'), reason="requires RCVTIMEO")
         def test_warn_set_timeo(self):
             s = self.context.socket(zmq.REQ)
             with warnings.catch_warnings(record=True) as w:
@@ -460,7 +557,7 @@ if have_gevent:
             self.assertEqual(w[0].category, UserWarning)
             
 
-        @skip_if(not hasattr(zmq, 'SNDTIMEO'))
+        @mark.skipif(not hasattr(zmq, 'SNDTIMEO'), reason="requires SNDTIMEO")
         def test_warn_get_timeo(self):
             s = self.context.socket(zmq.REQ)
             with warnings.catch_warnings(record=True) as w:
